"""
Phevo - an infectious game
Copyright (C) 2007  Mike Edwards

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

$Date: 2007-04-07 10:15:23 -0400 (Sat, 07 Apr 2007) $
$Revision: 25 $
"""
import random, config
from base import *
from engine import PygameEngine

random.seed(datetime.datetime.now().microsecond)

class Bug(Creature):

    def __init__(self,syringe,geno = None,mask=0,mutation_type=0,index=None):
        Creature.__init__(self,geno)
        self.syringe = syringe
        self.mask = mask
        self.ticks = 0
        self.mutation_type = mutation_type
        self.index = index
        self.notify("created")

    def __repr__(self):
        output = ""
        for i in range(1,-1,-1):
            if (self.mask >> (i * 2)) & 3 == 3:
                output += " %s" % config.debug_names[i][self.pheno[1-i%2]]
        return output
        
    def test(self,phevo):
        if not phevo.population:
            print "Population missing"
        if phevo.population:
            if self.mask != 0 and ((phevo.geno & self.mask) == (self.geno & self.mask)) and self.syringe.player != phevo.population.player:
                self.debug("phevo: %d, bug: %d, mask: %d" % (phevo.geno, self.geno, self.mask))
                return True
        return False

    def create_color_mask(self):
        mask = self.mask
        
        if (mask >> 3) & 3 != 0:
            mask &= 12
            if mask != 0:
                self.mask = mask
                
    def create_shape_mask(self):
        mask = self.mask
        
        if (mask >> 3) & 3 != 0:
            mask &= 3
            if mask != 0:
                self.mask = mask
                
    def on_syringe_mutation_requested(self,syringe):
        if self.syringe == syringe:
            #self.debug("Mask = %d" % self.mask)
            self.notify("mutated")
                
    def update(self):
        self.ticks += 1
        self.ticks %= config.bug_mutation_period
        
        if self.ticks == 0 and self.mutation_type > 0:
            if self.mutation_type == 1:
                self.create_color_mask()
            elif self.mutation_type == 2:
                self.create_shape_mask()
                
            self.notify("mutated")

class Syringe(Agent):
    def __init__(self, player, bugs = None):
        Agent.__init__(self)
        self.player = player
        if not bugs:
            self.bugs = []
        else:
            self.bugs = bugs
        self.mask = config.mask
        self.ticks = 0

    def spray(self,target,index = None):
        if target.player != self.player:
            if len(self.bugs) > 0:
                bug = self.bugs.pop(0)
                self.notify("sprayed",target,bug,index)
                self.debug("sprayed")
            
            #self.fill()
#        if len(self.bugs) == 0:
#            self.notify("emptied")

    def fill(self,phevo = None,mutation_type=0):
        if not phevo:
            phevo = random.choice(self.player.pop.well.keys())
        bug = Bug(self,phevo.geno,15,mutation_type,phevo.index)
        #bug.mask = bug.create_random_mask(False)
        self.bugs.append(bug)
        self.notify("filled",bug)
        phevo.promote()
        
    def update(self):
        self.ticks += 1
        self.ticks %= config.bug_mutation_period
        
        for bug in self.bugs:
            bug.update()
                #bug.mask = bug.create_random_mask(False)
                #bug.create_stronger_mask()
                #self.notify("mutation_requested")
        

class Phevo(Creature):

    def __init__(self, geno = None, population = None, index = 0):
        super(Phevo,self).__init__(geno)
        self.population = population
        self.index = index
        self.infected = False
        self.dead = False
        self.promoted = False
        self.neighbors = {}
        self.bug = None

    def __repr__(self):
        return "[%d] %s %s" % (self.index,\
                                config.debug_names[1][self.pheno[0]], \
                                config.debug_names[0][self.pheno[1]])

    def breed(self, mate, allow_mutants = True):
        mask = self.create_random_mask()
        offspring = Phevo(mask & self.geno | ~mask & mate.geno)
        if allow_mutants and random.randint(1,config.mutation_rate) == 1:
            offspring.geno ^= 1 << random.randint(0,3)
            self.debug("Mutation!")
        return offspring

    def check(self):
        if self.infected:
            self.cough()
            if random.randint(1,config.cure_rate) == 1:
                self.cure()
            if random.randint(1,config.kill_rate) == 1:
                self.die()

    def connect_neighbor(self,neighbor, dir=0):
        if neighbor:
            if not self.neighbors.has_key(neighbor):
                self.neighbors[neighbor] = dir
            if not neighbor.neighbors.has_key(self):
                neighbor.neighbors[self] = (dir + config.max_neighbors/2) % config.max_neighbors

    def disconnect_neighbor(self,neighbor):
        if self.neighbors.has_key(neighbor):
            self.neighbors.pop(neighbor)
        if neighbor.neighbors.has_key(self):
            neighbor.neighbors.pop(self)

    def disconnect_neighbors(self):
        for neighbor in self.neighbors.keys():
            if neighbor.neighbors.has_key(self):
                neighbor.neighbors.pop(self)
        self.neighbors = {}

    def infect(self,bug):
        self.debug("%s infected with %s" % (self,bug))
        self.infected = True
        self.bug = bug
        self.notify("infected")        

    def cure(self):
        self.infected = False
        self.notify("cured")

    def die(self):
        self.infected = False
        self.dead = True
        self.notify("died")
        self.disconnect_neighbors()
        self.bug = None
        self.population = None
        
    def promote(self):
        self.infected = False
        self.dead = False
        self.promoted = True
        self.notify("promoted")
        self.disconnect_neighbors()
        self.bug = None
        self.population = None
        
    def on_bug_attacked(self,bug,phevo):
        if self==phevo:
            self.infect(bug)

    def on_bug_cured(self,bug,phevo):
        if self==phevo:
            self.cure()

    def on_phevopop_check_requested(self,population):
        if population == self.population:
            self.check()

    def cough(self):
        self.notify("coughed")
        for phevo in self.neighbors.keys():
            if not phevo.infected \
               and not phevo.dead \
               and not phevo.bug == self.bug \
               and self.bug.test(phevo) \
               and random.randint(1,config.infect_rate) == 1: #immunity
                phevo.infect(self.bug)
        
class PhevoPop(Agent):
    MINIMUM = 16
    def __init__(self, player, phevos = None):
        Agent.__init__(self)
        self.player = player
        if phevos:
            self.phevos = phevos
        else:
            self.phevos = []
        self.sick = {}
        self.well = {}
        self.incubating = []
        self.this_update = 0
        self.last_update = 0

    def populate(self, phevos = None, generations = 6):
        self.phevos = []
        
        for index in range(2**generations):
            phevo = Phevo()
            self.phevos.append(phevo)
            phevo.index = index
            phevo.population = self
            self.well[phevo] = None
            phevo.notify("created")
            
        """
        generations -= 1
        if not phevos:
            phevos = []
            phevos.append(Phevo())
            phevos.append(Phevo())
        while generations > 0:
            new_pop = []
            for m in range(2):
                for x in range(len(phevos)):
                    new_pop.append(phevos[x].breed(phevos[x + (x % 2) * -2 + 1]))
            phevos = new_pop
            return self.populate(phevos, generations)
        if generations == 0:
            self.phevos = phevos
            random.shuffle(self.phevos)
            for index,phevo in enumerate(self.phevos):
                phevo.index = index
                phevo.population = self
                self.well[phevo] = None
                #phevo.subscribe(self)
                phevo.notify("created")
        """


    def connect(self):
        height, width = (8,8)
        for y in range(height):
            for x in range(width):
                index = x + width * y
                """
                Directions:
                  0
                3 + 1
                  2
                """
                if x % width != 0:
                    self.phevos[index].connect_neighbor(self.phevos[index - 1],1)
                #if x == width:
                #    self.phevos[index].connect_neighbor(self.phevos[-1 * width])
                    
                if y > 0:
                    self.phevos[index].connect_neighbor(self.phevos[index - 1 * width],2)
                #if y == height:
                #    self.phevos[index].connect_neighbor(self.phevos[x])
                """                    
                if x % width != 0 and y > 0:
                    subjects[-1].subscribe(subjects[-1 * width - 2])
                if x % width != width-1 and y > 0:
                    subjects[-1].subscribe(subjects[-1 * width - 0])
                """

    def update(self):
        self.this_update = self.player.ticks
        if self.player.ticks % config.population_update_rate == 0:
            self.notify("check_requested")
        if self.this_update - self.last_update > config.population_rebirth_rate and len(self.incubating) > 0:
            for i in range(len(self.incubating)):
                self.rebirth(self.incubating.pop())
            self.last_update = self.this_update
        if len(self.well) + len(self.sick) <= PhevoPop.MINIMUM:
            self.notify("empty")
        #print self.this_update - self.last_update, len(self.well), len(self.sick), len(self.incubating), len(self.well) + len(self.sick) + len(self.incubating)

    def remove_phevo(self,phevo):
        if self.sick.has_key(phevo):
            self.sick.pop(phevo)
        if self.well.has_key(phevo):
            self.well.pop(phevo)
        if phevo.population == self:
            self.phevos[phevo.index] = None
            
    def on_syringe_sprayed(self,syringe,target,bug,index = None):
        if target == self and len(self.well) > 0:
            self.debug("invaded")
            if index:
                phevo = self.phevos[index]
                if phevo:
                    if phevo.population == self:
                        if bug.test(phevo):
                            bug.notify("attacked",phevo)
            else:
                attempts = 0
                while attempts < len(self.well):
                    phevo = random.choice(self.well.keys())
                    if phevo.population == self:
                        if bug.test(phevo):
                            bug.notify("attacked",phevo)
                            break
                    attempts += 1
                    #else:
                    #    bug.notify("released",phevo)
        else:
            index = bug.index
            self.incubating.append(index)
            
    def rebirth(self,index):
        x = index % config.grid_width
        y = index / config.grid_height
        
        self.debug("Player %d Phevo %d promoted" % (self.player.id + 1, index))
        new_phevo = random.choice(self.well.keys()).breed(random.choice(self.well.keys()))
        
        phevo=new_phevo #random.choice(self.well.keys()).breed(random.choice(self.well.keys()))
        self.phevos[index] = phevo
        phevo.index = index
        phevo.population = self
        #self.well[phevo] = None
        
        self.debug("Connecting to (clockwise) %d %d %d %d" % (phevo.index - 1,phevo.index + 1,phevo.index - 1 * config.grid_width,phevo.index + 1 * config.grid_width))
        if x != 0:
            phevo.connect_neighbor(self.phevos[phevo.index - 1],1)
        if x < config.grid_width - 1:
            phevo.connect_neighbor(self.phevos[phevo.index + 1],1)
        if y > 0:
            phevo.connect_neighbor(self.phevos[phevo.index - 1 * config.grid_width],2)
        if y < config.grid_height - 1:
            phevo.connect_neighbor(self.phevos[phevo.index + 1 * config.grid_width],2)
        
        if len(phevo.neighbors) > 0:
            new_phevo=random.choice(phevo.neighbors.keys()).breed(random.choice(phevo.neighbors.keys()),False)
            neighbors = phevo.neighbors
            index = phevo.index
            #self.well.pop(phevo)
            phevo = new_phevo
            self.phevos[index] = phevo
            phevo.neighbors = neighbors
            phevo.index = index
            phevo.population = self
            self.debug("Redefined")

        if phevo.population == self:
            self.well[phevo] = None

        phevo.notify("created")

    def on_phevo_created(self,phevo):
        if phevo.population == self:
            self.notify("populated",phevo)

    def on_phevo_infected(self,phevo):
        if self.well.has_key(phevo):
            self.well.pop(phevo)
        if phevo.population == self:
            self.sick[phevo] = None

    def on_phevo_cured(self,phevo):
        if self.sick.has_key(phevo):
            self.sick.pop(phevo)
        if phevo.population == self:
            self.well[phevo] = None

    def on_phevo_died(self,phevo):
        self.remove_phevo(phevo)
            
    def on_phevo_promoted(self,phevo):
        #if phevo.population == self:
        self.remove_phevo(phevo)
        """
        if phevo.population == self:
            index = phevo.index
            self.debug("Player %d Phevo %d promoted" % (phevo.population.player.id + 1, index))
            if len(phevo.neighbors) > 0:
                new_phevo=random.choice(phevo.neighbors.keys()).breed(random.choice(phevo.neighbors.keys()))
            else:
                new_phevo = random.choice(self.well.keys()).breed(random.choice(self.well.keys()))
            self.remove_phevo(phevo)
            phevo=new_phevo #random.choice(self.well.keys()).breed(random.choice(self.well.keys()))
            self.phevos[index] = phevo
            phevo.index = index
            phevo.population = self
            self.well[phevo] = None
            phevo.notify("created")
            
            x = phevo.index % config.grid_width
            y = phevo.index / config.grid_height
            
            self.debug("Connecting to (clockwise) %d %d %d %d" % (phevo.index - 1,phevo.index + 1,phevo.index - 1 * config.grid_width,phevo.index + 1 * config.grid_width))
            if x != 0:
                phevo.connect_neighbor(self.phevos[phevo.index - 1],1)
            if x < config.grid_width - 1:
                phevo.connect_neighbor(self.phevos[phevo.index + 1],1)
            if y > 0:
                phevo.connect_neighbor(self.phevos[phevo.index - 1 * config.grid_width],2)
            if y < config.grid_height - 1:
                phevo.connect_neighbor(self.phevos[phevo.index + 1 * config.grid_width],2)
        """
        
 
            

class Player(Agent):
    def __init__(self,app,id,local = True, AI = True):
        Agent.__init__(self)
        self.app = app
        self.id = id
        self.pop = PhevoPop(self)
        self.syringe = Syringe(self)
        self.local = local
        self.AI = AI
        
        self.opponents = {}
        
        self.pop.populate()
        self.pop.connect()

        self.ticks = 0L
        
        #for i in range(2):
        #    self.syringe.fill()

    def add_opponent(self,player):
        if not self.opponents.has_key(player):
            self.opponents[player] = 0
        if not player.opponents.has_key(self):
            player.opponents[self] = 0
        
    def sub_opponent(self,player):
        if self.opponents.has_key(player):
            self.opponents.pop(player)
        if player.opponents.has_key(self):
            player.opponents.pop(self)

    def on_pygameengine_ticked(self,engine):
        self.ticks +=1
        self.pop.update()
        self.syringe.update()

    """
    def on_pygameengine_spray_requested(self,engine,id):
        if id == self.id:
            self.debug("spraying")
            #this is where you'll need to add the choice of phevo
            self.syringe.fill()
            for opponent in self.opponents.keys():
                self.syringe.spray(opponent.pop)
    """
        
    def on_cursor_promotion_located(self,cursor,id,index,mutation_type=0):
        if id == self.id and index < len(self.pop.phevos):
            if not self.pop.phevos[index] == None and len(self.syringe.bugs) < config.max_syringe_bugs:
                self.debug("promoting")
                #this is where you'll need to add the choice of phevo
                self.syringe.fill(self.pop.phevos[index],mutation_type)
            #elif self.pop.phevos[index] == None:
            #    pass
            elif self.pop.phevos[index] == None:
                self.debug("rotten phevo: player %s " % (id+1))
            
    def on_pygameengine_spray_requested(self,engine,id,index = None):
        if self.id == id:
            for opponent in self.opponents.keys():
                self.debug("hitting opponent")
                self.syringe.spray(opponent.pop,index)
        
    def on_phevopop_empty(self,pop):
        if self == pop.player:
            self.notify("defeated")
    #def on_syringe_emptied(self,syringe):
    #    if syringe == self.syringe:
    #        syringe.fill()

class App(Agent):
    def __init__(self):
        Agent.__init__(self)
        self.engine = None
        self.players = range(config.players)
            #self.players[player_id].syringe.fill()

    def start(self):
        for player_id in range(len(self.players)):
            if player_id > 0:
                self.players[player_id] = Player(self,player_id,AI = True)
                self.players[player_id].add_opponent(self.players[player_id - 1])
            else:
                self.players[player_id] = Player(self,player_id,AI = False)

        self.engine.start()

        if self.engine:
            #main loop
            while not(self.engine.finished):
                self.engine.tick()

    def stop(self):
        if self.engine:
            self.engine.stop()
        
if __name__ == "__main__":
    app = App()
    engine = PygameEngine(app,config.width,config.height)

    try:
        app.start()
    finally:
        app.stop()    
